/*
 * Copyright © 2017 Rob Clark <robclark@freedesktop.org>
 * SPDX-License-Identifier: MIT
 *
 * Authors:
 *    Rob Clark <robclark@freedesktop.org>
 */

#ifndef FD5_COMPUTE_H_
#define FD5_COMPUTE_H_

#include "pipe/p_context.h"

void fd5_compute_init(struct pipe_context *pctx);

#endif /* FD5_COMPUTE_H_ */
