wait(1)         n0<1>UD                                         { align1 WE_all 1N };
wait(1)         n0.1<1>UD                                       { align1 WE_all 1N };
wait(1)         n0.2<1>UD                                       { align1 WE_all 1N };
