/*
 * SPDX-FileCopyrightText: Copyright (c) 2001-2010 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: MIT
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/* AUTO GENERATED FILE -- DO NOT EDIT */

#ifndef __CLA097TEX_H__
#define __CLA097TEX_H__

/*
** Texture Header State
 */

#define NVA097_TEXHEAD0_COMPONENT_SIZES                            5:0
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R32_G32_B32_A32            0x00000001
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R32_G32_B32                0x00000002
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R16_G16_B16_A16            0x00000003
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R32_G32                    0x00000004
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R32_B24G8                  0x00000005
#define NVA097_TEXHEAD0_COMPONENT_SIZES_X8B8G8R8                   0x00000007
#define NVA097_TEXHEAD0_COMPONENT_SIZES_A8B8G8R8                   0x00000008
#define NVA097_TEXHEAD0_COMPONENT_SIZES_A2B10G10R10                0x00000009
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R16_G16                    0x0000000c
#define NVA097_TEXHEAD0_COMPONENT_SIZES_G8R24                      0x0000000d
#define NVA097_TEXHEAD0_COMPONENT_SIZES_G24R8                      0x0000000e
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R32                        0x0000000f
#define NVA097_TEXHEAD0_COMPONENT_SIZES_A4B4G4R4                   0x00000012
#define NVA097_TEXHEAD0_COMPONENT_SIZES_A5B5G5R1                   0x00000013
#define NVA097_TEXHEAD0_COMPONENT_SIZES_A1B5G5R5                   0x00000014
#define NVA097_TEXHEAD0_COMPONENT_SIZES_B5G6R5                     0x00000015
#define NVA097_TEXHEAD0_COMPONENT_SIZES_B6G5R5                     0x00000016
#define NVA097_TEXHEAD0_COMPONENT_SIZES_G8R8                       0x00000018
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R16                        0x0000001b
#define NVA097_TEXHEAD0_COMPONENT_SIZES_Y8_VIDEO                   0x0000001c
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R8                         0x0000001d
#define NVA097_TEXHEAD0_COMPONENT_SIZES_G4R4                       0x0000001e
#define NVA097_TEXHEAD0_COMPONENT_SIZES_R1                         0x0000001f
#define NVA097_TEXHEAD0_COMPONENT_SIZES_E5B9G9R9_SHAREDEXP         0x00000020
#define NVA097_TEXHEAD0_COMPONENT_SIZES_BF10GF11RF11               0x00000021
#define NVA097_TEXHEAD0_COMPONENT_SIZES_G8B8G8R8                   0x00000022
#define NVA097_TEXHEAD0_COMPONENT_SIZES_B8G8R8G8                   0x00000023
#define NVA097_TEXHEAD0_COMPONENT_SIZES_DXT1                       0x00000024
#define NVA097_TEXHEAD0_COMPONENT_SIZES_DXT23                      0x00000025
#define NVA097_TEXHEAD0_COMPONENT_SIZES_DXT45                      0x00000026
#define NVA097_TEXHEAD0_COMPONENT_SIZES_DXN1                       0x00000027
#define NVA097_TEXHEAD0_COMPONENT_SIZES_DXN2                       0x00000028
#define NVA097_TEXHEAD0_COMPONENT_SIZES_BC6H_SF16                  0x00000010
#define NVA097_TEXHEAD0_COMPONENT_SIZES_BC6H_UF16                  0x00000011
#define NVA097_TEXHEAD0_COMPONENT_SIZES_BC7U                       0x00000017
#define NVA097_TEXHEAD0_COMPONENT_SIZES_Z24S8                      0x00000029
#define NVA097_TEXHEAD0_COMPONENT_SIZES_X8Z24                      0x0000002a
#define NVA097_TEXHEAD0_COMPONENT_SIZES_S8Z24                      0x0000002b
#define NVA097_TEXHEAD0_COMPONENT_SIZES_X4V4Z24__COV4R4V           0x0000002c
#define NVA097_TEXHEAD0_COMPONENT_SIZES_X4V4Z24__COV8R8V           0x0000002d
#define NVA097_TEXHEAD0_COMPONENT_SIZES_V8Z24__COV4R12V            0x0000002e
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32                       0x0000002f
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X24S8                 0x00000030
#define NVA097_TEXHEAD0_COMPONENT_SIZES_X8Z24_X20V4S8__COV4R4V     0x00000031
#define NVA097_TEXHEAD0_COMPONENT_SIZES_X8Z24_X20V4S8__COV8R8V     0x00000032
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X20V4X8__COV4R4V      0x00000033
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X20V4X8__COV8R8V      0x00000034
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X20V4S8__COV4R4V      0x00000035
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X20V4S8__COV8R8V      0x00000036
#define NVA097_TEXHEAD0_COMPONENT_SIZES_X8Z24_X16V8S8__COV4R12V    0x00000037
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X16V8X8__COV4R12V     0x00000038
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X16V8S8__COV4R12V     0x00000039
#define NVA097_TEXHEAD0_COMPONENT_SIZES_Z16                        0x0000003a
#define NVA097_TEXHEAD0_COMPONENT_SIZES_V8Z24__COV8R24V            0x0000003b
#define NVA097_TEXHEAD0_COMPONENT_SIZES_X8Z24_X16V8S8__COV8R24V    0x0000003c
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X16V8X8__COV8R24V     0x0000003d
#define NVA097_TEXHEAD0_COMPONENT_SIZES_ZF32_X16V8S8__COV8R24V     0x0000003e
#define NVA097_TEXHEAD0_COMPONENT_SIZES_CS_BITFIELD_SIZE           0x0000003f
#define NVA097_TEXHEAD0_R_DATA_TYPE                                8:6
#define NVA097_TEXHEAD0_R_DATA_TYPE_NUM_SNORM                      0x00000001
#define NVA097_TEXHEAD0_R_DATA_TYPE_NUM_UNORM                      0x00000002
#define NVA097_TEXHEAD0_R_DATA_TYPE_NUM_SINT                       0x00000003
#define NVA097_TEXHEAD0_R_DATA_TYPE_NUM_UINT                       0x00000004
#define NVA097_TEXHEAD0_R_DATA_TYPE_NUM_SNORM_FORCE_FP16           0x00000005
#define NVA097_TEXHEAD0_R_DATA_TYPE_NUM_UNORM_FORCE_FP16           0x00000006
#define NVA097_TEXHEAD0_R_DATA_TYPE_NUM_FLOAT                      0x00000007
#define NVA097_TEXHEAD0_G_DATA_TYPE                                11:9
#define NVA097_TEXHEAD0_G_DATA_TYPE_NUM_SNORM                      0x00000001
#define NVA097_TEXHEAD0_G_DATA_TYPE_NUM_UNORM                      0x00000002
#define NVA097_TEXHEAD0_G_DATA_TYPE_NUM_SINT                       0x00000003
#define NVA097_TEXHEAD0_G_DATA_TYPE_NUM_UINT                       0x00000004
#define NVA097_TEXHEAD0_G_DATA_TYPE_NUM_SNORM_FORCE_FP16           0x00000005
#define NVA097_TEXHEAD0_G_DATA_TYPE_NUM_UNORM_FORCE_FP16           0x00000006
#define NVA097_TEXHEAD0_G_DATA_TYPE_NUM_FLOAT                      0x00000007
#define NVA097_TEXHEAD0_B_DATA_TYPE                                14:12
#define NVA097_TEXHEAD0_B_DATA_TYPE_NUM_SNORM                      0x00000001
#define NVA097_TEXHEAD0_B_DATA_TYPE_NUM_UNORM                      0x00000002
#define NVA097_TEXHEAD0_B_DATA_TYPE_NUM_SINT                       0x00000003
#define NVA097_TEXHEAD0_B_DATA_TYPE_NUM_UINT                       0x00000004
#define NVA097_TEXHEAD0_B_DATA_TYPE_NUM_SNORM_FORCE_FP16           0x00000005
#define NVA097_TEXHEAD0_B_DATA_TYPE_NUM_UNORM_FORCE_FP16           0x00000006
#define NVA097_TEXHEAD0_B_DATA_TYPE_NUM_FLOAT                      0x00000007
#define NVA097_TEXHEAD0_A_DATA_TYPE                                17:15
#define NVA097_TEXHEAD0_A_DATA_TYPE_NUM_SNORM                      0x00000001
#define NVA097_TEXHEAD0_A_DATA_TYPE_NUM_UNORM                      0x00000002
#define NVA097_TEXHEAD0_A_DATA_TYPE_NUM_SINT                       0x00000003
#define NVA097_TEXHEAD0_A_DATA_TYPE_NUM_UINT                       0x00000004
#define NVA097_TEXHEAD0_A_DATA_TYPE_NUM_SNORM_FORCE_FP16           0x00000005
#define NVA097_TEXHEAD0_A_DATA_TYPE_NUM_UNORM_FORCE_FP16           0x00000006
#define NVA097_TEXHEAD0_A_DATA_TYPE_NUM_FLOAT                      0x00000007
#define NVA097_TEXHEAD0_X_SOURCE                                   20:18
#define NVA097_TEXHEAD0_X_SOURCE_IN_ZERO                           0x00000000
#define NVA097_TEXHEAD0_X_SOURCE_IN_R                              0x00000002
#define NVA097_TEXHEAD0_X_SOURCE_IN_G                              0x00000003
#define NVA097_TEXHEAD0_X_SOURCE_IN_B                              0x00000004
#define NVA097_TEXHEAD0_X_SOURCE_IN_A                              0x00000005
#define NVA097_TEXHEAD0_X_SOURCE_IN_ONE_INT                        0x00000006
#define NVA097_TEXHEAD0_X_SOURCE_IN_ONE_FLOAT                      0x00000007
#define NVA097_TEXHEAD0_Y_SOURCE                                   23:21
#define NVA097_TEXHEAD0_Y_SOURCE_IN_ZERO                           0x00000000
#define NVA097_TEXHEAD0_Y_SOURCE_IN_R                              0x00000002
#define NVA097_TEXHEAD0_Y_SOURCE_IN_G                              0x00000003
#define NVA097_TEXHEAD0_Y_SOURCE_IN_B                              0x00000004
#define NVA097_TEXHEAD0_Y_SOURCE_IN_A                              0x00000005
#define NVA097_TEXHEAD0_Y_SOURCE_IN_ONE_INT                        0x00000006
#define NVA097_TEXHEAD0_Y_SOURCE_IN_ONE_FLOAT                      0x00000007
#define NVA097_TEXHEAD0_Z_SOURCE                                   26:24
#define NVA097_TEXHEAD0_Z_SOURCE_IN_ZERO                           0x00000000
#define NVA097_TEXHEAD0_Z_SOURCE_IN_R                              0x00000002
#define NVA097_TEXHEAD0_Z_SOURCE_IN_G                              0x00000003
#define NVA097_TEXHEAD0_Z_SOURCE_IN_B                              0x00000004
#define NVA097_TEXHEAD0_Z_SOURCE_IN_A                              0x00000005
#define NVA097_TEXHEAD0_Z_SOURCE_IN_ONE_INT                        0x00000006
#define NVA097_TEXHEAD0_Z_SOURCE_IN_ONE_FLOAT                      0x00000007
#define NVA097_TEXHEAD0_W_SOURCE                                   29:27
#define NVA097_TEXHEAD0_W_SOURCE_IN_ZERO                           0x00000000
#define NVA097_TEXHEAD0_W_SOURCE_IN_R                              0x00000002
#define NVA097_TEXHEAD0_W_SOURCE_IN_G                              0x00000003
#define NVA097_TEXHEAD0_W_SOURCE_IN_B                              0x00000004
#define NVA097_TEXHEAD0_W_SOURCE_IN_A                              0x00000005
#define NVA097_TEXHEAD0_W_SOURCE_IN_ONE_INT                        0x00000006
#define NVA097_TEXHEAD0_W_SOURCE_IN_ONE_FLOAT                      0x00000007
#define NVA097_TEXHEAD0_PACK_COMPONENTS                            30:30
#define NVA097_TEXHEAD0_RESERVED4                                  31:31
#define NVA097_TEXHEAD1_OFFSET_LOWER                               31:0
#define NVA097_TEXHEAD2_OFFSET_UPPER                               7:0
#define NVA097_TEXHEAD2_ANISO_SPREAD_MAX_LOG2_L_S_B                9:8
#define NVA097_TEXHEAD2_S_R_G_B_CONVERSION                         10:10
#define NVA097_TEXHEAD2_ANISO_SPREAD_MAX_LOG2_M_S_B                11:11
#define NVA097_TEXHEAD2_LOD_ANISO_QUALITY2                         12:12
#define NVA097_TEXHEAD2_COLOR_KEY_OP                               13:13
#define NVA097_TEXHEAD2_TEXTURE_TYPE                               17:14
#define NVA097_TEXHEAD2_TEXTURE_TYPE_ONE_D                         0x00000000
#define NVA097_TEXHEAD2_TEXTURE_TYPE_TWO_D                         0x00000001
#define NVA097_TEXHEAD2_TEXTURE_TYPE_THREE_D                       0x00000002
#define NVA097_TEXHEAD2_TEXTURE_TYPE_CUBEMAP                       0x00000003
#define NVA097_TEXHEAD2_TEXTURE_TYPE_ONE_D_ARRAY                   0x00000004
#define NVA097_TEXHEAD2_TEXTURE_TYPE_TWO_D_ARRAY                   0x00000005
#define NVA097_TEXHEAD2_TEXTURE_TYPE_ONE_D_BUFFER                  0x00000006
#define NVA097_TEXHEAD2_TEXTURE_TYPE_TWO_D_NO_MIPMAP               0x00000007
#define NVA097_TEXHEAD2_TEXTURE_TYPE_CUBEMAP_ARRAY                 0x00000008
#define NVA097_TEXHEAD2_TEXTURE_TYPE_TT_BIT_FIELD_SIZE             0x0000000f
#define NVA097_TEXHEAD2_MEMORY_LAYOUT                              18:18
#define NVA097_TEXHEAD2_MEMORY_LAYOUT_BLOCKLINEAR                  0x00000000
#define NVA097_TEXHEAD2_MEMORY_LAYOUT_PITCH                        0x00000001
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_WIDTH                       21:19
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_WIDTH_ONE_GOB               0x00000000
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_HEIGHT                      24:22
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_HEIGHT_ONE_GOB              0x00000000
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_HEIGHT_TWO_GOBS             0x00000001
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_HEIGHT_FOUR_GOBS            0x00000002
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_HEIGHT_EIGHT_GOBS           0x00000003
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_HEIGHT_SIXTEEN_GOBS         0x00000004
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_HEIGHT_THIRTYTWO_GOBS       0x00000005
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_DEPTH                       27:25
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_DEPTH_ONE_GOB               0x00000000
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_DEPTH_TWO_GOBS              0x00000001
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_DEPTH_FOUR_GOBS             0x00000002
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_DEPTH_EIGHT_GOBS            0x00000003
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_DEPTH_SIXTEEN_GOBS          0x00000004
#define NVA097_TEXHEAD2_GOBS_PER_BLOCK_DEPTH_THIRTYTWO_GOBS        0x00000005
#define NVA097_TEXHEAD2_SECTOR_PROMOTION                           29:28
#define NVA097_TEXHEAD2_SECTOR_PROMOTION_NO_PROMOTION              0x00000000
#define NVA097_TEXHEAD2_SECTOR_PROMOTION_PROMOTE_TO_2_V            0x00000001
#define NVA097_TEXHEAD2_SECTOR_PROMOTION_PROMOTE_TO_2_H            0x00000002
#define NVA097_TEXHEAD2_SECTOR_PROMOTION_PROMOTE_TO_4              0x00000003
#define NVA097_TEXHEAD2_BORDER_SOURCE                              30:30
#define NVA097_TEXHEAD2_BORDER_SOURCE_BORDER_TEXTURE               0x00000000
#define NVA097_TEXHEAD2_BORDER_SOURCE_BORDER_COLOR                 0x00000001
#define NVA097_TEXHEAD2_NORMALIZED_COORDS                          31:31
#define NVA097_TEXHEAD3_PITCH                                      19:0
#define NVA097_TEXHEAD3_LOD_ANISO_QUALITY                          20:20
#define NVA097_TEXHEAD3_LOD_ANISO_QUALITY_LOD_QUALITY_LOW          0x00000000
#define NVA097_TEXHEAD3_LOD_ANISO_QUALITY_LOD_QUALITY_HIGH         0x00000001
#define NVA097_TEXHEAD3_LOD_ISO_QUALITY                            21:21
#define NVA097_TEXHEAD3_LOD_ISO_QUALITY_LOD_QUALITY_LOW            0x00000000
#define NVA097_TEXHEAD3_LOD_ISO_QUALITY_LOD_QUALITY_HIGH           0x00000001
#define NVA097_TEXHEAD3_ANISO_COARSE_SPREAD_MODIFIER               23:22
#define NVA097_TEXHEAD3_ANISO_COARSE_SPREAD_MODIFIER_SPREAD_MODIFIER_NONE 0x00000000
#define NVA097_TEXHEAD3_ANISO_COARSE_SPREAD_MODIFIER_SPREAD_MODIFIER_CONST_ONE 0x00000001
#define NVA097_TEXHEAD3_ANISO_COARSE_SPREAD_MODIFIER_SPREAD_MODIFIER_CONST_TWO 0x00000002
#define NVA097_TEXHEAD3_ANISO_COARSE_SPREAD_MODIFIER_SPREAD_MODIFIER_SQRT 0x00000003
#define NVA097_TEXHEAD3_ANISO_SPREAD_SCALE                         28:24
#define NVA097_TEXHEAD3_USE_HEADER_OPT_CONTROL                     29:29
#define NVA097_TEXHEAD3_RESERVED3A                                 30:30
#define NVA097_TEXHEAD3_RESERVED3B                                 31:31
#define NVA097_TEXHEAD4_WIDTH                                      29:0
#define NVA097_TEXHEAD4_DEPTH_TEXTURE                              30:30
#define NVA097_TEXHEAD4_USE_TEXTURE_HEADER_VERSION2                31:31
#define NVA097_TEXHEAD5_HEIGHT                                     15:0
#define NVA097_TEXHEAD5_DEPTH                                      27:16
#define NVA097_TEXHEAD5_MAX_MIP_LEVEL                              31:28
#define NVA097_TEXHEAD6_TRILIN_OPT                                 4:0
#define NVA097_TEXHEAD6_MIP_LOD_BIAS                               17:5
#define NVA097_TEXHEAD6_RESERVED6A                                 18:18
#define NVA097_TEXHEAD6_ANISO_BIAS                                 22:19
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_FUNC                     24:23
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_FUNC_SPREAD_FUNC_HALF    0x00000000
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_FUNC_SPREAD_FUNC_ONE     0x00000001
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_FUNC_SPREAD_FUNC_TWO     0x00000002
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_FUNC_SPREAD_FUNC_MAX     0x00000003
#define NVA097_TEXHEAD6_ANISO_COARSE_SPREAD_FUNC                   26:25
#define NVA097_TEXHEAD6_ANISO_COARSE_SPREAD_FUNC_SPREAD_FUNC_HALF  0x00000000
#define NVA097_TEXHEAD6_ANISO_COARSE_SPREAD_FUNC_SPREAD_FUNC_ONE   0x00000001
#define NVA097_TEXHEAD6_ANISO_COARSE_SPREAD_FUNC_SPREAD_FUNC_TWO   0x00000002
#define NVA097_TEXHEAD6_ANISO_COARSE_SPREAD_FUNC_SPREAD_FUNC_MAX   0x00000003
#define NVA097_TEXHEAD6_MAX_ANISOTROPY                             29:27
#define NVA097_TEXHEAD6_MAX_ANISOTROPY_ANISO_1_TO_1                0x00000000
#define NVA097_TEXHEAD6_MAX_ANISOTROPY_ANISO_2_TO_1                0x00000001
#define NVA097_TEXHEAD6_MAX_ANISOTROPY_ANISO_4_TO_1                0x00000002
#define NVA097_TEXHEAD6_MAX_ANISOTROPY_ANISO_6_TO_1                0x00000003
#define NVA097_TEXHEAD6_MAX_ANISOTROPY_ANISO_8_TO_1                0x00000004
#define NVA097_TEXHEAD6_MAX_ANISOTROPY_ANISO_10_TO_1               0x00000005
#define NVA097_TEXHEAD6_MAX_ANISOTROPY_ANISO_12_TO_1               0x00000006
#define NVA097_TEXHEAD6_MAX_ANISOTROPY_ANISO_16_TO_1               0x00000007
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_MODIFIER                 31:30
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_MODIFIER_SPREAD_MODIFIER_NONE 0x00000000
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_MODIFIER_SPREAD_MODIFIER_CONST_ONE 0x00000001
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_MODIFIER_SPREAD_MODIFIER_CONST_TWO 0x00000002
#define NVA097_TEXHEAD6_ANISO_FINE_SPREAD_MODIFIER_SPREAD_MODIFIER_SQRT 0x00000003
#define NVA097_TEXHEAD7_COLOR_KEY_VALUE                            31:0


/*
** Texture Header State, Version 2
 */

#define NVA097_TEXHEADV2_0_COMPONENT_SIZES                         5:0
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R32_G32_B32_A32         0x00000001
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R32_G32_B32             0x00000002
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R16_G16_B16_A16         0x00000003
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R32_G32                 0x00000004
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R32_B24G8               0x00000005
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_X8B8G8R8                0x00000007
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_A8B8G8R8                0x00000008
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_A2B10G10R10             0x00000009
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R16_G16                 0x0000000c
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_G8R24                   0x0000000d
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_G24R8                   0x0000000e
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R32                     0x0000000f
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_A4B4G4R4                0x00000012
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_A5B5G5R1                0x00000013
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_A1B5G5R5                0x00000014
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_B5G6R5                  0x00000015
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_B6G5R5                  0x00000016
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_G8R8                    0x00000018
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R16                     0x0000001b
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_Y8_VIDEO                0x0000001c
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R8                      0x0000001d
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_G4R4                    0x0000001e
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_R1                      0x0000001f
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_E5B9G9R9_SHAREDEXP      0x00000020
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_BF10GF11RF11            0x00000021
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_G8B8G8R8                0x00000022
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_B8G8R8G8                0x00000023
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_DXT1                    0x00000024
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_DXT23                   0x00000025
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_DXT45                   0x00000026
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_DXN1                    0x00000027
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_DXN2                    0x00000028
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_BC6H_SF16               0x00000010
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_BC6H_UF16               0x00000011
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_BC7U                    0x00000017
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_Z24S8                   0x00000029
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_X8Z24                   0x0000002a
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_S8Z24                   0x0000002b
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_X4V4Z24__COV4R4V        0x0000002c
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_X4V4Z24__COV8R8V        0x0000002d
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_V8Z24__COV4R12V         0x0000002e
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32                    0x0000002f
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X24S8              0x00000030
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_X8Z24_X20V4S8__COV4R4V  0x00000031
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_X8Z24_X20V4S8__COV8R8V  0x00000032
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X20V4X8__COV4R4V   0x00000033
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X20V4X8__COV8R8V   0x00000034
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X20V4S8__COV4R4V   0x00000035
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X20V4S8__COV8R8V   0x00000036
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_X8Z24_X16V8S8__COV4R12V 0x00000037
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X16V8X8__COV4R12V  0x00000038
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X16V8S8__COV4R12V  0x00000039
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_Z16                     0x0000003a
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_V8Z24__COV8R24V         0x0000003b
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_X8Z24_X16V8S8__COV8R24V 0x0000003c
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X16V8X8__COV8R24V  0x0000003d
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_ZF32_X16V8S8__COV8R24V  0x0000003e
#define NVA097_TEXHEADV2_0_COMPONENT_SIZES_CS_BITFIELD_SIZE        0x0000003f
#define NVA097_TEXHEADV2_0_R_DATA_TYPE                             8:6
#define NVA097_TEXHEADV2_0_R_DATA_TYPE_NUM_SNORM                   0x00000001
#define NVA097_TEXHEADV2_0_R_DATA_TYPE_NUM_UNORM                   0x00000002
#define NVA097_TEXHEADV2_0_R_DATA_TYPE_NUM_SINT                    0x00000003
#define NVA097_TEXHEADV2_0_R_DATA_TYPE_NUM_UINT                    0x00000004
#define NVA097_TEXHEADV2_0_R_DATA_TYPE_NUM_SNORM_FORCE_FP16        0x00000005
#define NVA097_TEXHEADV2_0_R_DATA_TYPE_NUM_UNORM_FORCE_FP16        0x00000006
#define NVA097_TEXHEADV2_0_R_DATA_TYPE_NUM_FLOAT                   0x00000007
#define NVA097_TEXHEADV2_0_G_DATA_TYPE                             11:9
#define NVA097_TEXHEADV2_0_G_DATA_TYPE_NUM_SNORM                   0x00000001
#define NVA097_TEXHEADV2_0_G_DATA_TYPE_NUM_UNORM                   0x00000002
#define NVA097_TEXHEADV2_0_G_DATA_TYPE_NUM_SINT                    0x00000003
#define NVA097_TEXHEADV2_0_G_DATA_TYPE_NUM_UINT                    0x00000004
#define NVA097_TEXHEADV2_0_G_DATA_TYPE_NUM_SNORM_FORCE_FP16        0x00000005
#define NVA097_TEXHEADV2_0_G_DATA_TYPE_NUM_UNORM_FORCE_FP16        0x00000006
#define NVA097_TEXHEADV2_0_G_DATA_TYPE_NUM_FLOAT                   0x00000007
#define NVA097_TEXHEADV2_0_B_DATA_TYPE                             14:12
#define NVA097_TEXHEADV2_0_B_DATA_TYPE_NUM_SNORM                   0x00000001
#define NVA097_TEXHEADV2_0_B_DATA_TYPE_NUM_UNORM                   0x00000002
#define NVA097_TEXHEADV2_0_B_DATA_TYPE_NUM_SINT                    0x00000003
#define NVA097_TEXHEADV2_0_B_DATA_TYPE_NUM_UINT                    0x00000004
#define NVA097_TEXHEADV2_0_B_DATA_TYPE_NUM_SNORM_FORCE_FP16        0x00000005
#define NVA097_TEXHEADV2_0_B_DATA_TYPE_NUM_UNORM_FORCE_FP16        0x00000006
#define NVA097_TEXHEADV2_0_B_DATA_TYPE_NUM_FLOAT                   0x00000007
#define NVA097_TEXHEADV2_0_A_DATA_TYPE                             17:15
#define NVA097_TEXHEADV2_0_A_DATA_TYPE_NUM_SNORM                   0x00000001
#define NVA097_TEXHEADV2_0_A_DATA_TYPE_NUM_UNORM                   0x00000002
#define NVA097_TEXHEADV2_0_A_DATA_TYPE_NUM_SINT                    0x00000003
#define NVA097_TEXHEADV2_0_A_DATA_TYPE_NUM_UINT                    0x00000004
#define NVA097_TEXHEADV2_0_A_DATA_TYPE_NUM_SNORM_FORCE_FP16        0x00000005
#define NVA097_TEXHEADV2_0_A_DATA_TYPE_NUM_UNORM_FORCE_FP16        0x00000006
#define NVA097_TEXHEADV2_0_A_DATA_TYPE_NUM_FLOAT                   0x00000007
#define NVA097_TEXHEADV2_0_X_SOURCE                                20:18
#define NVA097_TEXHEADV2_0_X_SOURCE_IN_ZERO                        0x00000000
#define NVA097_TEXHEADV2_0_X_SOURCE_IN_R                           0x00000002
#define NVA097_TEXHEADV2_0_X_SOURCE_IN_G                           0x00000003
#define NVA097_TEXHEADV2_0_X_SOURCE_IN_B                           0x00000004
#define NVA097_TEXHEADV2_0_X_SOURCE_IN_A                           0x00000005
#define NVA097_TEXHEADV2_0_X_SOURCE_IN_ONE_INT                     0x00000006
#define NVA097_TEXHEADV2_0_X_SOURCE_IN_ONE_FLOAT                   0x00000007
#define NVA097_TEXHEADV2_0_Y_SOURCE                                23:21
#define NVA097_TEXHEADV2_0_Y_SOURCE_IN_ZERO                        0x00000000
#define NVA097_TEXHEADV2_0_Y_SOURCE_IN_R                           0x00000002
#define NVA097_TEXHEADV2_0_Y_SOURCE_IN_G                           0x00000003
#define NVA097_TEXHEADV2_0_Y_SOURCE_IN_B                           0x00000004
#define NVA097_TEXHEADV2_0_Y_SOURCE_IN_A                           0x00000005
#define NVA097_TEXHEADV2_0_Y_SOURCE_IN_ONE_INT                     0x00000006
#define NVA097_TEXHEADV2_0_Y_SOURCE_IN_ONE_FLOAT                   0x00000007
#define NVA097_TEXHEADV2_0_Z_SOURCE                                26:24
#define NVA097_TEXHEADV2_0_Z_SOURCE_IN_ZERO                        0x00000000
#define NVA097_TEXHEADV2_0_Z_SOURCE_IN_R                           0x00000002
#define NVA097_TEXHEADV2_0_Z_SOURCE_IN_G                           0x00000003
#define NVA097_TEXHEADV2_0_Z_SOURCE_IN_B                           0x00000004
#define NVA097_TEXHEADV2_0_Z_SOURCE_IN_A                           0x00000005
#define NVA097_TEXHEADV2_0_Z_SOURCE_IN_ONE_INT                     0x00000006
#define NVA097_TEXHEADV2_0_Z_SOURCE_IN_ONE_FLOAT                   0x00000007
#define NVA097_TEXHEADV2_0_W_SOURCE                                29:27
#define NVA097_TEXHEADV2_0_W_SOURCE_IN_ZERO                        0x00000000
#define NVA097_TEXHEADV2_0_W_SOURCE_IN_R                           0x00000002
#define NVA097_TEXHEADV2_0_W_SOURCE_IN_G                           0x00000003
#define NVA097_TEXHEADV2_0_W_SOURCE_IN_B                           0x00000004
#define NVA097_TEXHEADV2_0_W_SOURCE_IN_A                           0x00000005
#define NVA097_TEXHEADV2_0_W_SOURCE_IN_ONE_INT                     0x00000006
#define NVA097_TEXHEADV2_0_W_SOURCE_IN_ONE_FLOAT                   0x00000007
#define NVA097_TEXHEADV2_0_PACK_COMPONENTS                         30:30
#define NVA097_TEXHEADV2_0_RESERVED4                               31:31
#define NVA097_TEXHEADV2_1_OFFSET_LOWER                            31:0
#define NVA097_TEXHEADV2_2_OFFSET_UPPER                            7:0
#define NVA097_TEXHEADV2_2_ANISO_SPREAD_MAX_LOG2_L_S_B             9:8
#define NVA097_TEXHEADV2_2_S_R_G_B_CONVERSION                      10:10
#define NVA097_TEXHEADV2_2_ANISO_SPREAD_MAX_LOG2_M_S_B             11:11
#define NVA097_TEXHEADV2_2_LOD_ANISO_QUALITY2                      12:12
#define NVA097_TEXHEADV2_2_COLOR_KEY_OP                            13:13
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE                            17:14
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_ONE_D                      0x00000000
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_TWO_D                      0x00000001
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_THREE_D                    0x00000002
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_CUBEMAP                    0x00000003
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_ONE_D_ARRAY                0x00000004
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_TWO_D_ARRAY                0x00000005
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_ONE_D_BUFFER               0x00000006
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_TWO_D_NO_MIPMAP            0x00000007
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_CUBEMAP_ARRAY              0x00000008
#define NVA097_TEXHEADV2_2_TEXTURE_TYPE_TT_BIT_FIELD_SIZE          0x0000000f
#define NVA097_TEXHEADV2_2_MEMORY_LAYOUT                           18:18
#define NVA097_TEXHEADV2_2_MEMORY_LAYOUT_BLOCKLINEAR               0x00000000
#define NVA097_TEXHEADV2_2_MEMORY_LAYOUT_PITCH                     0x00000001
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_WIDTH                    21:19
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_WIDTH_ONE_GOB            0x00000000
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_HEIGHT                   24:22
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_HEIGHT_ONE_GOB           0x00000000
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_HEIGHT_TWO_GOBS          0x00000001
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_HEIGHT_FOUR_GOBS         0x00000002
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_HEIGHT_EIGHT_GOBS        0x00000003
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_HEIGHT_SIXTEEN_GOBS      0x00000004
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_HEIGHT_THIRTYTWO_GOBS    0x00000005
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_DEPTH                    27:25
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_DEPTH_ONE_GOB            0x00000000
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_DEPTH_TWO_GOBS           0x00000001
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_DEPTH_FOUR_GOBS          0x00000002
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_DEPTH_EIGHT_GOBS         0x00000003
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_DEPTH_SIXTEEN_GOBS       0x00000004
#define NVA097_TEXHEADV2_2_GOBS_PER_BLOCK_DEPTH_THIRTYTWO_GOBS     0x00000005
#define NVA097_TEXHEADV2_2_SECTOR_PROMOTION                        29:28
#define NVA097_TEXHEADV2_2_SECTOR_PROMOTION_NO_PROMOTION           0x00000000
#define NVA097_TEXHEADV2_2_SECTOR_PROMOTION_PROMOTE_TO_2_V         0x00000001
#define NVA097_TEXHEADV2_2_SECTOR_PROMOTION_PROMOTE_TO_2_H         0x00000002
#define NVA097_TEXHEADV2_2_SECTOR_PROMOTION_PROMOTE_TO_4           0x00000003
#define NVA097_TEXHEADV2_2_BORDER_SOURCE                           30:30
#define NVA097_TEXHEADV2_2_BORDER_SOURCE_BORDER_TEXTURE            0x00000000
#define NVA097_TEXHEADV2_2_BORDER_SOURCE_BORDER_COLOR              0x00000001
#define NVA097_TEXHEADV2_2_NORMALIZED_COORDS                       31:31
#define NVA097_TEXHEADV2_3_PITCH                                   19:0
#define NVA097_TEXHEADV2_3_LOD_ANISO_QUALITY                       20:20
#define NVA097_TEXHEADV2_3_LOD_ANISO_QUALITY_LOD_QUALITY_LOW       0x00000000
#define NVA097_TEXHEADV2_3_LOD_ANISO_QUALITY_LOD_QUALITY_HIGH      0x00000001
#define NVA097_TEXHEADV2_3_LOD_ISO_QUALITY                         21:21
#define NVA097_TEXHEADV2_3_LOD_ISO_QUALITY_LOD_QUALITY_LOW         0x00000000
#define NVA097_TEXHEADV2_3_LOD_ISO_QUALITY_LOD_QUALITY_HIGH        0x00000001
#define NVA097_TEXHEADV2_3_ANISO_COARSE_SPREAD_MODIFIER            23:22
#define NVA097_TEXHEADV2_3_ANISO_COARSE_SPREAD_MODIFIER_SPREAD_MODIFIER_NONE 0x00000000
#define NVA097_TEXHEADV2_3_ANISO_COARSE_SPREAD_MODIFIER_SPREAD_MODIFIER_CONST_ONE 0x00000001
#define NVA097_TEXHEADV2_3_ANISO_COARSE_SPREAD_MODIFIER_SPREAD_MODIFIER_CONST_TWO 0x00000002
#define NVA097_TEXHEADV2_3_ANISO_COARSE_SPREAD_MODIFIER_SPREAD_MODIFIER_SQRT 0x00000003
#define NVA097_TEXHEADV2_3_ANISO_SPREAD_SCALE                      28:24
#define NVA097_TEXHEADV2_3_USE_HEADER_OPT_CONTROL                  29:29
#define NVA097_TEXHEADV2_3_RESERVED3A                              30:30
#define NVA097_TEXHEADV2_3_RESERVED3B                              31:31
#define NVA097_TEXHEADV2_4_WIDTH                                   29:0
#define NVA097_TEXHEADV2_4_DEPTH_TEXTURE                           30:30
#define NVA097_TEXHEADV2_4_USE_TEXTURE_HEADER_VERSION2             31:31
#define NVA097_TEXHEADV2_5_HEIGHT                                  15:0
#define NVA097_TEXHEADV2_5_DEPTH                                   27:16
#define NVA097_TEXHEADV2_5_MAX_MIP_LEVEL                           31:28
#define NVA097_TEXHEADV2_6_TRILIN_OPT                              4:0
#define NVA097_TEXHEADV2_6_MIP_LOD_BIAS                            17:5
#define NVA097_TEXHEADV2_6_RESERVED6A                              18:18
#define NVA097_TEXHEADV2_6_ANISO_BIAS                              22:19
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_FUNC                  24:23
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_FUNC_SPREAD_FUNC_HALF 0x00000000
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_FUNC_SPREAD_FUNC_ONE  0x00000001
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_FUNC_SPREAD_FUNC_TWO  0x00000002
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_FUNC_SPREAD_FUNC_MAX  0x00000003
#define NVA097_TEXHEADV2_6_ANISO_COARSE_SPREAD_FUNC                26:25
#define NVA097_TEXHEADV2_6_ANISO_COARSE_SPREAD_FUNC_SPREAD_FUNC_HALF 0x00000000
#define NVA097_TEXHEADV2_6_ANISO_COARSE_SPREAD_FUNC_SPREAD_FUNC_ONE 0x00000001
#define NVA097_TEXHEADV2_6_ANISO_COARSE_SPREAD_FUNC_SPREAD_FUNC_TWO 0x00000002
#define NVA097_TEXHEADV2_6_ANISO_COARSE_SPREAD_FUNC_SPREAD_FUNC_MAX 0x00000003
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY                          29:27
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY_ANISO_1_TO_1             0x00000000
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY_ANISO_2_TO_1             0x00000001
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY_ANISO_4_TO_1             0x00000002
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY_ANISO_6_TO_1             0x00000003
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY_ANISO_8_TO_1             0x00000004
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY_ANISO_10_TO_1            0x00000005
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY_ANISO_12_TO_1            0x00000006
#define NVA097_TEXHEADV2_6_MAX_ANISOTROPY_ANISO_16_TO_1            0x00000007
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_MODIFIER              31:30
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_MODIFIER_SPREAD_MODIFIER_NONE 0x00000000
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_MODIFIER_SPREAD_MODIFIER_CONST_ONE 0x00000001
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_MODIFIER_SPREAD_MODIFIER_CONST_TWO 0x00000002
#define NVA097_TEXHEADV2_6_ANISO_FINE_SPREAD_MODIFIER_SPREAD_MODIFIER_SQRT 0x00000003
#define NVA097_TEXHEADV2_7_RES_VIEW_MIN_MIP_LEVEL                  3:0
#define NVA097_TEXHEADV2_7_RES_VIEW_MAX_MIP_LEVEL                  7:4
#define NVA097_TEXHEADV2_7_HEIGHT_MSB                              8:8
#define NVA097_TEXHEADV2_7_HEIGHT_MSB_RESERVED                     11:9
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT                      15:12
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_1X1             0x00000000
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_2X1             0x00000001
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_2X2             0x00000002
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_4X2             0x00000003
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_4X2_D3D         0x00000004
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_2X1_D3D         0x00000005
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_4X4             0x00000006
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_2X2_VC_4        0x00000008
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_2X2_VC_12       0x00000009
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_4X2_VC_8        0x0000000a
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_4X2_VC_24       0x0000000b
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_2X1_CENTER      0x0000000c
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_2X2_CENTER      0x0000000d
#define NVA097_TEXHEADV2_7_MULTI_SAMPLE_COUNT_MODE_4X2_CENTER      0x0000000e
#define NVA097_TEXHEADV2_7_MIN_LOD_CLAMP                           27:16
#define NVA097_TEXHEADV2_7_DEPTH_MSB                               30:28
#define NVA097_TEXHEADV2_7_RESERVED7A                              31:31


/*
** Texture Sampler State
 */

#define NVA097_TEXSAMP0_ADDRESS_U                                  2:0
#define NVA097_TEXSAMP0_ADDRESS_U_WRAP                             0x00000000
#define NVA097_TEXSAMP0_ADDRESS_U_MIRROR                           0x00000001
#define NVA097_TEXSAMP0_ADDRESS_U_CLAMP_TO_EDGE                    0x00000002
#define NVA097_TEXSAMP0_ADDRESS_U_BORDER                           0x00000003
#define NVA097_TEXSAMP0_ADDRESS_U_CLAMP_OGL                        0x00000004
#define NVA097_TEXSAMP0_ADDRESS_U_MIRROR_ONCE_CLAMP_TO_EDGE        0x00000005
#define NVA097_TEXSAMP0_ADDRESS_U_MIRROR_ONCE_BORDER               0x00000006
#define NVA097_TEXSAMP0_ADDRESS_U_MIRROR_ONCE_CLAMP_OGL            0x00000007
#define NVA097_TEXSAMP0_ADDRESS_V                                  5:3
#define NVA097_TEXSAMP0_ADDRESS_V_WRAP                             0x00000000
#define NVA097_TEXSAMP0_ADDRESS_V_MIRROR                           0x00000001
#define NVA097_TEXSAMP0_ADDRESS_V_CLAMP_TO_EDGE                    0x00000002
#define NVA097_TEXSAMP0_ADDRESS_V_BORDER                           0x00000003
#define NVA097_TEXSAMP0_ADDRESS_V_CLAMP_OGL                        0x00000004
#define NVA097_TEXSAMP0_ADDRESS_V_MIRROR_ONCE_CLAMP_TO_EDGE        0x00000005
#define NVA097_TEXSAMP0_ADDRESS_V_MIRROR_ONCE_BORDER               0x00000006
#define NVA097_TEXSAMP0_ADDRESS_V_MIRROR_ONCE_CLAMP_OGL            0x00000007
#define NVA097_TEXSAMP0_ADDRESS_P                                  8:6
#define NVA097_TEXSAMP0_ADDRESS_P_WRAP                             0x00000000
#define NVA097_TEXSAMP0_ADDRESS_P_MIRROR                           0x00000001
#define NVA097_TEXSAMP0_ADDRESS_P_CLAMP_TO_EDGE                    0x00000002
#define NVA097_TEXSAMP0_ADDRESS_P_BORDER                           0x00000003
#define NVA097_TEXSAMP0_ADDRESS_P_CLAMP_OGL                        0x00000004
#define NVA097_TEXSAMP0_ADDRESS_P_MIRROR_ONCE_CLAMP_TO_EDGE        0x00000005
#define NVA097_TEXSAMP0_ADDRESS_P_MIRROR_ONCE_BORDER               0x00000006
#define NVA097_TEXSAMP0_ADDRESS_P_MIRROR_ONCE_CLAMP_OGL            0x00000007
#define NVA097_TEXSAMP0_DEPTH_COMPARE                              9:9
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC                         12:10
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC_ZC_NEVER                0x00000000
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC_ZC_LESS                 0x00000001
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC_ZC_EQUAL                0x00000002
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC_ZC_LEQUAL               0x00000003
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC_ZC_GREATER              0x00000004
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC_ZC_NOTEQUAL             0x00000005
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC_ZC_GEQUAL               0x00000006
#define NVA097_TEXSAMP0_DEPTH_COMPARE_FUNC_ZC_ALWAYS               0x00000007
#define NVA097_TEXSAMP0_S_R_G_B_CONVERSION                         13:13
#define NVA097_TEXSAMP0_RESERVED0A                                 16:14
#define NVA097_TEXSAMP0_RESERVED0B                                 19:17
#define NVA097_TEXSAMP0_MAX_ANISOTROPY                             22:20
#define NVA097_TEXSAMP0_MAX_ANISOTROPY_ANISO_1_TO_1                0x00000000
#define NVA097_TEXSAMP0_MAX_ANISOTROPY_ANISO_2_TO_1                0x00000001
#define NVA097_TEXSAMP0_MAX_ANISOTROPY_ANISO_4_TO_1                0x00000002
#define NVA097_TEXSAMP0_MAX_ANISOTROPY_ANISO_6_TO_1                0x00000003
#define NVA097_TEXSAMP0_MAX_ANISOTROPY_ANISO_8_TO_1                0x00000004
#define NVA097_TEXSAMP0_MAX_ANISOTROPY_ANISO_10_TO_1               0x00000005
#define NVA097_TEXSAMP0_MAX_ANISOTROPY_ANISO_12_TO_1               0x00000006
#define NVA097_TEXSAMP0_MAX_ANISOTROPY_ANISO_16_TO_1               0x00000007
#define NVA097_TEXSAMP1_MAG_FILTER                                 2:0
#define NVA097_TEXSAMP1_MAG_FILTER_MAG_POINT                       0x00000001
#define NVA097_TEXSAMP1_MAG_FILTER_MAG_LINEAR                      0x00000002
#define NVA097_TEXSAMP1_MAG_FILTER_VCAA_4_TAP                      0x00000003
#define NVA097_TEXSAMP1_MAG_FILTER_VCAA_8_TAP                      0x00000004
#define NVA097_TEXSAMP1_MIN_LOD_CLAMP_BEHAVIOR_FOR_NEAREST_MIP     3:3
#define NVA097_TEXSAMP1_MIN_LOD_CLAMP_BEHAVIOR_FOR_NEAREST_MIP_INTEGER_AND_FRACTION 0x00000000
#define NVA097_TEXSAMP1_MIN_LOD_CLAMP_BEHAVIOR_FOR_NEAREST_MIP_INTEGER_ONLY 0x00000001
#define NVA097_TEXSAMP1_MIN_FILTER                                 5:4
#define NVA097_TEXSAMP1_MIN_FILTER_MIN_POINT                       0x00000001
#define NVA097_TEXSAMP1_MIN_FILTER_MIN_LINEAR                      0x00000002
#define NVA097_TEXSAMP1_MIN_FILTER_MIN_ANISO                       0x00000003
#define NVA097_TEXSAMP1_MIP_FILTER                                 7:6
#define NVA097_TEXSAMP1_MIP_FILTER_MIP_NONE                        0x00000001
#define NVA097_TEXSAMP1_MIP_FILTER_MIP_POINT                       0x00000002
#define NVA097_TEXSAMP1_MIP_FILTER_MIP_LINEAR                      0x00000003
#define NVA097_TEXSAMP1_CUBEMAP_INTERFACE_FILTERING                9:8
#define NVA097_TEXSAMP1_CUBEMAP_INTERFACE_FILTERING_USE_WRAP       0x00000000
#define NVA097_TEXSAMP1_CUBEMAP_INTERFACE_FILTERING_OVERRIDE_WRAP  0x00000001
#define NVA097_TEXSAMP1_CUBEMAP_INTERFACE_FILTERING_AUTO_SPAN_SEAM 0x00000002
#define NVA097_TEXSAMP1_CUBEMAP_INTERFACE_FILTERING_AUTO_CROSS_SEAM 0x00000003
#define NVA097_TEXSAMP1_MIP_LOD_BIAS                               24:12
#define NVA097_TEXSAMP1_FLOAT_COORD_NORMALIZATION                  25:25
#define NVA097_TEXSAMP1_FLOAT_COORD_NORMALIZATION_USE_HEADER_SETTING 0x00000000
#define NVA097_TEXSAMP1_FLOAT_COORD_NORMALIZATION_FORCE_UNNORMALIZED_COORDS 0x00000001
#define NVA097_TEXSAMP1_TRILIN_OPT                                 30:26
#define NVA097_TEXSAMP2_MIN_LOD_CLAMP                              11:0
#define NVA097_TEXSAMP2_MAX_LOD_CLAMP                              23:12
#define NVA097_TEXSAMP2_S_R_G_B_BORDER_COLOR_R                     31:24
#define NVA097_TEXSAMP3_RESERVED12                                 11:0
#define NVA097_TEXSAMP3_S_R_G_B_BORDER_COLOR_G                     19:12
#define NVA097_TEXSAMP3_S_R_G_B_BORDER_COLOR_B                     27:20
#define NVA097_TEXSAMP4_BORDER_COLOR_R                             31:0
#define NVA097_TEXSAMP5_BORDER_COLOR_G                             31:0
#define NVA097_TEXSAMP6_BORDER_COLOR_B                             31:0
#define NVA097_TEXSAMP7_BORDER_COLOR_A                             31:0



#endif // #ifndef __CLA097TEX_H__
