/*
 * Copyright 2025 LunarG, Inc.
 * Copyright 2025 Google LLC
 * SPDX-License-Identifier: MIT
 */

#include "mtl_bridge.h"

mtl_texture *
mtl_drawable_get_texture(void *drawable_ptr)
{
   return NULL;
}

void *
mtl_retain(void *handle)
{
   return NULL;
}

void
mtl_release(void *handle)
{
}
