/*
 * Copyright © 2016 Rob Clark <robclark@freedesktop.org>
 * SPDX-License-Identifier: MIT
 *
 * Authors:
 *    Rob Clark <robclark@freedesktop.org>
 */

#ifndef FD5_QUERY_H_
#define FD5_QUERY_H_

#include "pipe/p_context.h"

void fd5_query_context_init(struct pipe_context *pctx);

#endif /* FD5_QUERY_H_ */
