/*
 * Copyright © 2024 Collabora Ltd.
 * SPDX-License-Identifier: MIT
 */

#ifndef NIL_RS_BINDINGS_H
#define NIL_RS_BINDINGS_H

#include "nil_format_table.h"
#include "util/format/u_format.h"
#include "nouveau/headers/nv_device_info.h"

#endif /* NIL_RS_BINDINGS_H */
