endif(8)        JIP: LABEL1                                          { align16 1Q };
LABEL1:
endif(8)        JIP: LABEL2                                          { align1 1Q };
LABEL2:
endif(16)       JIP: LABEL3                                          { align1 1H };
LABEL3:
